@@ -3,7 +3,7 @@
 # This file is part of Hypothesis, which may be found at
 # https://github.com/HypothesisWorks/hypothesis-python
 #
-# Most of this work is copyright (C) 2013-2016 David R. MacIver
+# Most of this work is copyright (C) 2013-2017 David R. MacIver
 # (david@drmaciver.com), but it contains contributions by others. See
 # CONTRIBUTING.rst for a full list of people who may hold copyright, and
 # consult the git log if you need to determine who owns an individual
@@ -25,9 +25,11 @@ from weakref import WeakKeyDictionary
 from hypothesis import settings as Settings
 from hypothesis import Phase
 from hypothesis.reporting import debug_report
-from hypothesis.internal.compat import hbytes, hrange, Counter, \
-    text_type, bytes_from_list, to_bytes_sequence, unicode_safe_repr
-from hypothesis.internal.conjecture.data import Status, StopTest, TestData
+from hypothesis.internal.compat import EMPTY_BYTES, Counter, hbytes, \
+    hrange, text_type, bytes_from_list, to_bytes_sequence, \
+    unicode_safe_repr
+from hypothesis.internal.conjecture.data import Status, StopTest, \
+    ConjectureData
 from hypothesis.internal.conjecture.minimizer import minimize
 
 
@@ -44,7 +46,7 @@ class RunIsComplete(Exception):
     pass
 
 
-class TestRunner(object):
+class ConjectureRunner(object):
 
     def __init__(
         self, test_function, settings=None, random=None,
@@ -61,16 +63,30 @@ class TestRunner(object):
         self.start_time = time.time()
         self.random = random or Random(getrandbits(128))
         self.database_key = database_key
-        self.seen = set()
-        self.duplicates = 0
         self.status_runtimes = {}
         self.events_to_strings = WeakKeyDictionary()
 
+        # Tree nodes are stored in an array to prevent heavy nesting of data
+        # structures. Branches are dicts mapping bytes to child nodes (which
+        # will in general only be partially populated). Leaves are
+        # ConjectureData objects that have been previously seen as the result
+        # of following that path.
+        self.tree = [{}]
+
+        # A node is dead if there is nothing left to explore past that point.
+        # Recursively, a node is dead if either it is a leaf or every byte
+        # leads to a dead node when starting from here.
+        self.dead = set()
+
+    def __tree_is_exhausted(self):
+        return 0 in self.dead
+
     def new_buffer(self):
-        self.last_data = TestData(
+        assert not self.__tree_is_exhausted()
+        self.last_data = ConjectureData(
             max_length=self.settings.buffer_size,
             draw_bytes=lambda data, n, distribution:
-            distribution(self.random, n)
+            self.__rewrite_for_novelty(data, distribution(self.random, n))
         )
         self.test_function(self.last_data)
         self.last_data.freeze()
@@ -90,27 +106,44 @@ class TestRunner(object):
         finally:
             data.freeze()
             self.note_details(data)
-        if (
-            data.status == Status.INTERESTING and (
-                self.last_data is None or
-                data.buffer != self.last_data.buffer
-            )
-        ):
-            self.debug_data(data)
+
+        self.debug_data(data)
         if data.status >= Status.VALID:
             self.valid_examples += 1
 
+        tree_node = self.tree[0]
+        indices = []
+        i = 0
+        for b in data.buffer:
+            indices.append(i)
+            try:
+                i = tree_node[b]
+            except KeyError:
+                i = len(self.tree)
+                self.tree.append({})
+                tree_node[b] = i
+            tree_node = self.tree[i]
+            if i in self.dead:
+                break
+
+        if data.status != Status.OVERRUN and i not in self.dead:
+            self.dead.add(i)
+            self.tree[i] = data
+
+            for j in reversed(indices):
+                if len(self.tree[j]) < 256:
+                    break
+                if set(self.tree[j].values()).issubset(self.dead):
+                    self.dead.add(j)
+                else:
+                    break
+
     def consider_new_test_data(self, data):
         # Transition rules:
         #   1. Transition cannot decrease the status
         #   2. Any transition which increases the status is valid
         #   3. If the previous status was interesting, only shrinking
         #      transitions are allowed.
-        key = hbytes(data.buffer)
-        if key in self.seen:
-            self.duplicates += 1
-            return False
-        self.seen.add(key)
         if data.buffer == self.last_data.buffer:
             return False
         if self.last_data.status < data.status:
@@ -159,8 +192,6 @@ class TestRunner(object):
         ))
 
     def incorporate_new_buffer(self, buffer):
-        if buffer in self.seen:
-            return False
         assert self.last_data.status == Status.INTERESTING
         if (
             self.settings.timeout > 0 and
@@ -168,11 +199,24 @@ class TestRunner(object):
         ):
             self.exit_reason = ExitReason.timeout
             raise RunIsComplete()
+
         buffer = buffer[:self.last_data.index]
         if sort_key(buffer) >= sort_key(self.last_data.buffer):
             return False
+
+        i = 0
+        for b in buffer:
+            if i in self.dead:
+                return False
+            try:
+                i = self.tree[i][b]
+            except KeyError:
+                break
+        else:
+            return False
+
         assert sort_key(buffer) <= sort_key(self.last_data.buffer)
-        data = TestData.for_buffer(buffer)
+        data = ConjectureData.for_buffer(buffer)
         self.test_function(data)
         if self.consider_new_test_data(data):
             self.shrinks += 1
@@ -224,7 +268,9 @@ class TestRunner(object):
                 i = self.random.choice(choices)
                 return self.last_data.buffer[i:i + n]
             else:
-                return distribution(self.random, n)
+                result = distribution(self.random, n)
+                assert isinstance(result, hbytes)
+                return result
 
         def flip_bit(data, n, distribution):
             buf = bytearray(
@@ -235,7 +281,7 @@ class TestRunner(object):
             return hbytes(buf)
 
         def draw_zero(data, n, distribution):
-            return b'\0' * n
+            return hbytes(b'\0' * n)
 
         def draw_constant(data, n, distribution):
             return bytes_from_list([
@@ -257,10 +303,64 @@ class TestRunner(object):
             if (
                 data.index + n > len(self.last_data.buffer)
             ):
-                return distribution(self.random, n)
-            return self.random.choice(bits)(data, n, distribution)
+                result = distribution(self.random, n)
+            else:
+                result = self.random.choice(bits)(data, n, distribution)
+
+            return self.__rewrite_for_novelty(data, result)
+
         return draw_mutated
 
+    def __rewrite_for_novelty(self, data, result):
+        assert isinstance(result, hbytes)
+        try:
+            node_index = data.__current_node_index
+        except AttributeError:
+            assert len(data.buffer) == 0
+            node_index = 0
+            data.__current_node_index = node_index
+            data.__hit_novelty = False
+
+        if data.__hit_novelty:
+            return result
+
+        node = self.tree[node_index]
+        assert node_index not in self.dead
+
+        for i, b in enumerate(result):
+            assert isinstance(b, int)
+            try:
+                new_node_index = node[b]
+            except KeyError:
+                data.__hit_novelty = True
+                return result
+
+            new_node = self.tree[new_node_index]
+
+            if new_node_index in self.dead:
+                if isinstance(result, hbytes):
+                    result = bytearray(result)
+                for c in range(256):
+                    if c not in node:
+                        result[i] = c
+                        data.__hit_novelty = True
+                        return hbytes(result)
+                    else:
+                        new_node_index = node[c]
+                        new_node = self.tree[new_node_index]
+                        if new_node_index not in self.dead:
+                            result[i] = c
+                            break
+                else:  # pragma: no cover
+                    assert False, (
+                        'Found a tree node which is live despite all its '
+                        'children being dead.')
+            node_index = new_node_index
+            node = new_node
+        assert node_index not in self.dead
+        data.__current_node_index = node_index
+        return hbytes(result)
+
     def _run(self):
         self.last_data = None
         mutations = 0
@@ -283,10 +383,11 @@ class TestRunner(object):
                 ):
                     self.exit_reason = ExitReason.max_iterations
                     return
-                data = TestData.for_buffer(existing)
+                data = ConjectureData.for_buffer(existing)
                 self.test_function(data)
                 data.freeze()
                 self.last_data = data
+                self.consider_new_test_data(data)
                 if data.status < Status.VALID:
                     self.settings.database.delete(
                         self.database_key, existing)
@@ -304,7 +405,10 @@ class TestRunner(object):
                     self.last_data = data
                     break
 
-        if Phase.generate in self.settings.phases:
+        if (
+            Phase.generate in self.settings.phases and not
+            self.__tree_is_exhausted()
+        ):
             if (
                 self.last_data is None or
                 self.last_data.status < Status.INTERESTING
@@ -312,7 +416,10 @@ class TestRunner(object):
                 self.new_buffer()
 
             mutator = self._new_mutator()
-            while self.last_data.status != Status.INTERESTING:
+            while (
+                self.last_data.status != Status.INTERESTING and
+                not self.__tree_is_exhausted()
+            ):
                 if self.valid_examples >= self.settings.max_examples:
                     self.exit_reason = ExitReason.max_examples
                     return
@@ -332,7 +439,7 @@ class TestRunner(object):
                     self.new_buffer()
                     mutator = self._new_mutator()
                 else:
-                    data = TestData(
+                    data = ConjectureData(
                         draw_bytes=mutator,
                         max_length=self.settings.buffer_size
                     )
@@ -348,6 +455,10 @@ class TestRunner(object):
 
                 mutations += 1
 
+        if self.__tree_is_exhausted():
+            self.exit_reason = ExitReason.finished
+            return
+
         data = self.last_data
         if data is None:
             self.exit_reason = ExitReason.finished
@@ -362,11 +473,7 @@ class TestRunner(object):
             self.exit_reason = ExitReason.finished
             return
 
-        if not self.last_data.buffer:
-            self.exit_reason = ExitReason.finished
-            return
-
-        data = TestData.for_buffer(self.last_data.buffer)
+        data = ConjectureData.for_buffer(self.last_data.buffer)
         self.test_function(data)
         if data.status != Status.INTERESTING:
             self.exit_reason = ExitReason.flaky
@@ -376,46 +483,56 @@ class TestRunner(object):
 
         while self.changed > change_counter:
             change_counter = self.changed
-            failed_deletes = 0
-            while self.last_data.intervals and failed_deletes < 10:
-                if self.random.randint(0, 1):
-                    u, v = self.random.choice(self.last_data.intervals)
-                else:
-                    n = len(self.last_data.buffer) - 1
-                    u, v = sorted((
-                        self.random.choice(self.last_data.intervals)
-                    ))
-                if (
-                    v < len(self.last_data.buffer)
-                ) and self.incorporate_new_buffer(
-                    self.last_data.buffer[:u] +
-                    self.last_data.buffer[v:]
-                ):
-                    failed_deletes = 0
-                else:
-                    failed_deletes += 1
-            i = 0
-            while i < len(self.last_data.intervals):
-                u, v = self.last_data.intervals[i]
-                if not self.incorporate_new_buffer(
-                    self.last_data.buffer[:u] +
-                    self.last_data.buffer[v:]
-                ):
-                    i += 1
-            i = 0
-            while i + 1 < len(self.last_data.buffer):
-                if not self.incorporate_new_buffer(
-                    self.last_data.buffer[:i] +
-                    self.last_data.buffer[i + 1:]
-                ):
-                    i += 1
+
+            self.debug('Structured interval deletes')
+
+            k = len(self.last_data.intervals) // 2
+            while k > 0:
+                i = 0
+                while i + k <= len(self.last_data.intervals):
+                    bitmask = [True] * len(self.last_data.buffer)
+
+                    for u, v in self.last_data.intervals[i:i + k]:
+                        for t in range(u, v):
+                            bitmask[t] = False
+
+                    u, v = self.last_data.intervals[i]
+                    if not self.incorporate_new_buffer(hbytes(
+                        b for b, v in zip(self.last_data.buffer, bitmask)
+                        if v
+                    )):
+                        i += k
+                k //= 2
+
+            if change_counter != self.changed:
+                self.debug('Restarting')
+                continue
+
+            self.debug('Bulk replacing blocks with simpler blocks')
             i = 0
             while i < len(self.last_data.blocks):
                 u, v = self.last_data.blocks[i]
                 buf = self.last_data.buffer
                 block = buf[u:v]
                 n = v - u
-                all_blocks = sorted(set([bytes(n)] + [
+
+                buffer = bytearray()
+                for r, s in self.last_data.blocks:
+                    if s - r == n and self.last_data.buffer[r:s] > block:
+                        buffer.extend(block)
+                    else:
+                        buffer.extend(self.last_data.buffer[r:s])
+                self.incorporate_new_buffer(hbytes(buffer))
+                i += 1
+
+            self.debug('Replacing individual blocks with simpler blocks')
+            i = 0
+            while i < len(self.last_data.blocks):
+                u, v = self.last_data.blocks[i]
+                buf = self.last_data.buffer
+                block = buf[u:v]
+                n = v - u
+                all_blocks = sorted(set([hbytes(n)] + [
                     buf[a:a + n]
                     for a in self.last_data.block_starts[n]
                 ]))
@@ -427,6 +544,7 @@ class TestRunner(object):
                         break
                 i += 1
 
+            self.debug('Simultaneous shrinking of duplicated blocks')
             block_counter = -1
             while block_counter < self.changed:
                 block_counter = self.changed
@@ -444,15 +562,26 @@ class TestRunner(object):
                     ]
 
                     def replace(b):
-                        return b''.join(
-                            bytes(b if c == block else c) for c in parts
-                        )
+                        return hbytes(EMPTY_BYTES.join(
+                            hbytes(b if c == block else c) for c in parts
+                        ))
                     minimize(
                         block,
                         lambda b: self.incorporate_new_buffer(replace(b)),
                         self.random
                     )
 
+            if change_counter != self.changed:
+                self.debug('Restarting')
+                continue
+
+            self.debug('Lexicographical minimization of whole buffer')
+            minimize(
+                self.last_data.buffer, self.incorporate_new_buffer,
+                cautious=True
+            )
+
+            self.debug('Shrinking of individual blocks')
             i = 0
             while i < len(self.last_data.blocks):
                 u, v = self.last_data.blocks[i]
@@ -465,24 +594,67 @@ class TestRunner(object):
                 )
                 i += 1
 
-            i = 0
-            alternatives = None
-            while i < len(self.last_data.intervals):
-                if alternatives is None:
-                    alternatives = sorted(set(
-                        self.last_data.buffer[u:v]
-                        for u, v in self.last_data.intervals), key=len)
-                u, v = self.last_data.intervals[i]
-                for a in alternatives:
-                    buf = self.last_data.buffer
-                    if (
-                        len(a) < v - u or
-                        (len(a) == (v - u) and a < buf[u:v])
-                    ):
-                        if self.incorporate_new_buffer(buf[:u] + a + buf[v:]):
-                            alternatives = None
+            if change_counter != self.changed:
+                self.debug('Restarting')
+                continue
+
+            self.debug('Reordering blocks')
+            block_lengths = sorted(self.last_data.block_starts, reverse=True)
+            for n in block_lengths:
+                i = 1
+                while i < len(self.last_data.block_starts.get(n, ())):
+                    j = i
+                    while j > 0:
+                        buf = self.last_data.buffer
+                        blocks = self.last_data.block_starts[n]
+                        a_start = blocks[j - 1]
+                        b_start = blocks[j]
+                        a = buf[a_start:a_start + n]
+                        b = buf[b_start:b_start + n]
+                        if a <= b:
                             break
-                i += 1
+                        swapped = (
+                            buf[:a_start] + b + buf[a_start + n:b_start] +
+                            a + buf[b_start + n:])
+                        assert len(swapped) == len(buf)
+                        assert swapped < buf
+                        if self.incorporate_new_buffer(swapped):
+                            j -= 1
+                        else:
+                            break
+                    i += 1
+
+            if change_counter != self.changed:
+                self.debug('Restarting')
+                continue
+
+            self.debug('Shuffling suffixes while shrinking %r' % (
+                self.last_data.bind_points,
+            ))
+            b = 0
+            while b < len(self.last_data.bind_points):
+                cutoff = sorted(self.last_data.bind_points)[b]
+
+                def test_value(prefix):
+                    for t in hrange(5):
+                        alphabet = {}
+                        for i, j in self.last_data.blocks[b:]:
+                            alphabet.setdefault(j - i, []).append((i, j))
+                        if t > 0:
+                            for v in alphabet.values():
+                                self.random.shuffle(v)
+                        buf = bytearray(prefix)
+                        for i, j in self.last_data.blocks[b:]:
+                            u, v = alphabet[j - i].pop()
+                            buf.extend(self.last_data.buffer[u:v])
+                        if self.incorporate_new_buffer(hbytes(buf)):
+                            return True
+                    return False
+                minimize(
+                    self.last_data.buffer[:cutoff], test_value, cautious=True
+                )
+                b += 1
+
         self.exit_reason = ExitReason.finished
 
     def event_to_string(self, event):
